%%%
%{
% This code is the implementation of the following paper. If you use this 
code, please cite the paper.
@inproceedings{hok_pooling,
    author = {Anoop Cherian and Piotr Koniusz and Stephen Gould},
    title = {Higher-order Pooling of CNN Features via Kernel Linearization
    for Action Recognition},
    booktitle = {Winter Conference on Computer Vision (WACV)},
    year = {2017}
}   
% This code is released on the BSD3 license.
% This code should not be used for non-commercial purposes. 
% The authors are not liable to any loss or damage caused by running this code.
% For any issues/bugs, please contact anoop.cherian@gmail.com.

Copyright (c) 2017, Anoop Cherian All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, 
this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or 
other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its contributors may 
be used to endorse or promote products derived from this software without specific 
prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE 
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.
%}
%
% pre-requisites to run this code:
% install liblinear 2.x from https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/multicore-liblinear/
% install tensor toolbox for higher-order SVD from http://au.mathworks.com/matlabcentral/fileexchange/25514-tp-tool
% install vl_feat from http://www.vlfeat.org/
%
% This implementation generates HOK descriptors on CNN features from fc8
% layer of a standard VGG-16 CNN fine-tuned on the JHMDB action recognition dataset split-1.
%
function demo()    
    addpath ../../../tools/svm/liblinear2/matlab; % you need liblinear 2.x if you want to do classification.
    addpath(genpath('../../../tools/_ten3/tptool/')); % you need to install TP tool for hosvd.

    % load data.
    load 'data/jhmdb_data.mat'; % loads fg_data_imflow, fg_data_imdata, labels, train_seq_ids, test_seq_ids.
  
    % generate hok descriptors
    fprintf('generating HOK descriptors for flow stream...\n');
    hok_flow = get_hok_descriptors(fg_data_imflow, 'flow', 48); % non-gradient                                          
    fprintf('generating HOK descriptors for RGB stream...\n');
    hok_rgb = get_hok_descriptors(fg_data_imdata,'rgb', 48);                       
                    
    feat = [(cell2mat(hok_flow)), (cell2mat(hok_rgb))];                                        
    
    training_feat = feat(train_seq_ids,:);                     
	test_feat = feat(test_seq_ids,:);                  
 	training_labels = all_labels(train_seq_ids); 
    test_labels = all_labels(test_seq_ids);   

    fprintf('training an SVM...\n');
    test_accuracy = train_and_test_svm(training_feat, training_labels, test_feat, test_labels);            
    fprintf('classification accuracy = %f\n', test_accuracy);
end
          
function test_accuracy = train_and_test_svm(train_data, train_labels, test_data, test_labels)              
    params = ' -q -c 0.1 -s 0 -n 12 ';        

    % you need to have multi-core liblinear 2.x
    model = train(train_labels', sparse(train_data), params);    
    [~,b,~] = predict(test_labels', sparse(test_data), model, '-b 1');                              
    test_accuracy = b(1);
end

function xx = get_hok_descriptors(xx, type, num_pivots)
    xx = cellfun(@(yy) bsxfun(@times, 1./((sum(abs(yy'),2)+eps)), yy'), xx, 'uniformoutput', false); % normalize each cnn prediction to sum to one.
    
    % generate a set of classifier score pivots
    try              
        % load the respective pivots.
        load(['./data/pivots_' type '_' num2str(num_pivots) '.mat']);
    catch
        % you need to have vl_feat installed. 
        yy = xx(randperm(length(xx), 600));
        X = cell2mat(yy);       
        [mm, cc, pp] = vl_gmm(X', num_pivots); % you need to install vlfeat.
        pivots.mean = mm; pivots.dcov = cc; pivots.prior=pp;        
        parsave(['./data/pivots_' type '_' num2str(num_pivots) '.mat'], 'pivots');
    end
            
    for t=1:length(xx)     
       if mod(t,20)==0, fprintf('.'); end
       xx{t} = generate_hok_descriptor(xx{t}, pivots);                      
    end    
    fprintf('\n');
end
